structure DeflateIO :> DeflateIO =
struct

	type byteio = int * DeflateBits.byte list * DeflateBits.byte DeflateOutput.container
	(*
		#1: bits that have been extracted from input bytes
		#2: unprocessed input bytes
		#3: output bytes
	*)
	
	exception Unexpected_end_of_stream
	exception Unaligned_byte_operation

	(* byte operations *)
	
	fun explode_byte b = DeflateBits.explode_byte_lsb_first 8 b
	
	(* get one more byte into bits *)
	
	fun from_input bytes = (0, bytes, DeflateOutput.empty)
	
	fun from_input_string str = from_input (map DeflateBits.char_to_byte (explode str))
	
	fun from_input_ints ints = from_input (map DeflateBits.int_to_byte ints)
	
	(* output *)

	fun output (_, _, result) = DeflateOutput.output result
	
	fun output_string io = implode (map DeflateBits.byte_to_char (output io))
	
	fun append_bytes (bitpos, bs, result) rs = (bitpos, bs, DeflateOutput.append_bytes result rs)
	
	fun append_byte (bitpos, bs, result) r = (bitpos, bs, DeflateOutput.append_byte result r)
	
	fun append_from_output (bitpos, bs, result) nbytes distance = 
			(bitpos, bs, DeflateOutput.append_from_output result nbytes distance)
		
	fun flush_ouput (bits, bs, result) = (bits, bs, DeflateOutput.empty)
			
	(* string dump *)
	
	fun to_string (io as (bitpos, bytes, result)) = 
		"(" ^ Int.toString bitpos ^ " bitpos, " 
		^ Int.toString (length bytes) ^ " in, "
		^ Int.toString (length (output io)) ^ " out)"

	(* take/drop bytes from input *)

	fun take_bytes (0, bytes, _) nbytes = List.take (bytes, nbytes)
	  | take_bytes _ _ = raise Unaligned_byte_operation

	fun drop_bytes (0, bytes, result) nbytes = (0, List.drop (bytes, nbytes), result)
	  | drop_bytes _ _ = raise Unaligned_byte_operation

	fun take_zero_string (0, bytes, _) =
	  	let
			fun impl [] = []
			  | impl (b::bs) = 
			  		if DeflateBits.is_byte_zero b then
						[]
					else
						b :: impl bs
		in
			implode (map DeflateBits.byte_to_char (impl bytes))
		end
	  | take_zero_string _ = raise Unaligned_byte_operation
	  
	fun drop_zero_string (0, b::bs, result) =
	  		if DeflateBits.is_byte_zero b then
				(0, bs, result)
			else
				drop_zero_string (0, bs, result)
	  | drop_zero_string (io as (0, [], _)) = io
	  | drop_zero_string _ = raise Unaligned_byte_operation
	
	(* take/drop bits from input *)
	
	fun take_bit (bitpos, byte::_, _) = DeflateBits.byte_bit bitpos byte
	  | take_bit _ = raise Unexpected_end_of_stream

	fun drop_bit (_, [], _) = raise Unexpected_end_of_stream
	  | drop_bit (7, _::bytes, result) = (0, bytes, result)
	  | drop_bit (bitpos, bytes, result) = (bitpos + 1, bytes, result)
	  
	fun take_bits _ 0 = []
	  | take_bits io n = take_bit io :: take_bits (drop_bit io) (n - 1)
	   
	fun drop_bits io 0 = io
	  | drop_bits (bitpos, byte::bytes, result) nbits =
			let
				val rembits = nbits - (8 - bitpos)
		  	in
				if rembits > 0 then
					drop_bits 
						(0, List.drop (bytes, Int.quot (rembits, 8)), result)
						(Int.rem (rembits, 8)) 
				else if rembits = 0 then
					(0, bytes, result)
				else
					(bitpos + nbits, byte::bytes, result)
			end	
	  | drop_bits _ _ = raise Unexpected_end_of_stream

	fun drop_to_byte (io as (0, _, _)) = io
	  | drop_to_byte (_, _::bytes, result) = (0, bytes, result)
	  | drop_to_byte _ = raise Unexpected_end_of_stream
	
	(* get fixed size value *)
	fun take_fixed _ 0 = 0
	  | take_fixed (bitpos, byte::bytes, result) nbits =
			let
				val rembits = nbits - (8 - bitpos)
		  	in
				if rembits > 0 then
					DeflateBits.byte_sub (bitpos, 7) byte  
						+ (take_fixed (0, bytes, result) rembits * DeflateBits.pow2 (8 - bitpos))
				else
					DeflateBits.byte_sub (bitpos, bitpos + nbits - 1) byte
	  		end
	  | take_fixed _ _ = raise Unexpected_end_of_stream
	  
	fun get_fixed n io = (take_fixed io n, drop_bits io n)

	(* self tests *)
	fun eq_io (bitpos1, (b1::bytes1), out1) (bitpos2, (b2::bytes2), out2) =
			bitpos1 = bitpos2 
				andalso DeflateBits.byte_sub (bitpos1, 7) b1 = DeflateBits.byte_sub (bitpos2, 7) b2
				andalso List.all (fn x=>x) (ListPair.map DeflateBits.byte_eq (bytes1, bytes2))
				andalso eq_io (0,[],out1) (0,[],out2)
	  | eq_io (_,[],out1) (_,[],out2) = List.all (fn x=>x) (ListPair.map DeflateBits.byte_eq 
								(DeflateOutput.output out1, DeflateOutput.output out2))
	  | eq_io _ _ = false

	fun test_io (bits, ints, out_ints) : byteio = (8 - length bits,  
				map DeflateBits.int_to_byte 
				(DeflateBits.to_int_lsb_first bits * DeflateBits.pow2 (8 - length bits) :: ints),
				DeflateOutput.append_bytes DeflateOutput.empty (map DeflateBits.int_to_byte out_ints))

	val tests = [
		take_bit (test_io ([true],[],[])) = true,
		take_bit (test_io ([false],[],[])) = false,
		take_bit (test_io ([true,true,false,true],[],[])) = true,
		take_bit (test_io ([false,false,true,true],[],[])) = false,
		take_bit (test_io ([true,false,true,false,true,true,false,true],[],[])) = true,
		take_bit (test_io ([false,true,true,true,false,true,true,true],[],[])) = false,
		take_bits (test_io ([true,true,false], [2,4,8],[])) 2 = [true,true],
		take_bits (test_io ([false,true], [17,1,2],[])) 6 = [false,true,true,false,false,false],
		take_fixed (test_io ([false,true,true,true], [1,2,3],[4,5])) 3 = 6,
		take_fixed (test_io ([false,true], [33,1,2,3],[4,5])) 6 = 6,
		take_fixed (test_io ([false,true,true,true],[],[])) 2 = 2,
		eq_io (drop_bits (test_io ([true,true,false], [2,4,8],[2,1])) 2)
			(test_io ([false],[2,4,8],[2,1])),
		eq_io (drop_bits (test_io ([false,true], [17,1,2],[])) 6) 
			(test_io ([true,false,false,false],[1,2],[])),
		eq_io (drop_bits (test_io ([false,true], [17,1,2],[222])) 10) 
			(test_io ([true,false,false,false,false,false,false,false],[2],[222])),
		eq_io (drop_bits (test_io ([false,true], [17,2,2],[222])) 11) 
			(test_io ([true,false,false,false,false,false,false],[2],[222]))
		]

end
